import mx.ODBC.Windows as odbc
import asyncore
import sys

class QueryLogger:
	def __init__ (self, logger = None):
		self.logger = logger
	
	def __call__ (self, catch, query):
		self.log (catch, query)
				
	def log (self, catch, query):
		if self.logger:
			self.logger.log ("%s %s" % (catch, query))

class QueryError (Exception): pass		
class DBServerError (Exception): pass
		
class odbcf:
	def __init__ (self, conninfo, logger = None, log_query = 1):	
		self.conninfo = conninfo
		self.logger = logger
		self.log_query = log_query
		self.quqeylogger = QueryLogger (logger)
		self.conn = None
		self.cx = None
		self.connected = 0
	
	def __getattr__ (self, attr):
		if self.cx:
			return getattr (self.cx, attr)
		else:
			raise AttributeError	
	
	def log (self, line):	
		if self.logger:
			self.logger.log (line)
		else:
			print(line)	
	
	def trace (self, name = ""):
		if self.logger:
			self.logger.trace (name)
		else:
			print(asyncore.compact_traceback ())
				
	#----------------------------------------------------------------
	# database connect
	#----------------------------------------------------------------	
	def connect (self):
		connopt = self.conninfo.split ("/")
		self.conn = odbc.connect (*tuple (connopt [:3]))
		self.cx = self.conn.cursor ()
		
		if len (connopt) == 4:
			try:
				self.cx.execute ("use %s" % connopt [-1])
			except odbc.Warning:
				pass
			except:
				self.c.close ()
				self.conn.close ()
				raise
		
		self.execute ("set ANSI_NULLS ON; SET ANSI_WARNINGS ON; SET XACT_ABORT ON")
		self.commit ()		
		self.connected = 1
		
	def reconnect (self):
		self.close ()
		self.connect ()
		
	def close (self):
		if not self.connected: return
		
		if self.cx:
			self.execute ("set ANSI_NULLS OFF; SET ANSI_WARNINGS OFF; SET XACT_ABORT OFF")			
			self.commit ()
			try: self.cx.close ()
			except: pass
			self.cx = None
		
		if self.conn: 
			try: self.conn.close ()
			except: pass
			self.cx = None
			
		self.connected = 0
	
	def commit (self):
		self.conn.commit ()
	
	def rollback (self):
		self.conn.rollback ()
	
	def execute (self, query):
		try:
			self.cx.execute (query)			
		
		except odbc.OperationalError:
			raise DBServerError("mxODBC.OperationalError, %s %s %s %s" % tuple (why [:4]))
			
		except Exception as why:
			errstr = "%s %s %s %s" % tuple (why [:4])
			if self.log_query: self.quqeylogger (errstr, query)
			try: raise
			except: what = sys.exc_info()[0]
			raise QueryError("%s, %s" % (what, errstr))
	


import threading
from collections import deque

class ready_producer_fifo:
	# On http2, it is impossible to bind channel.ready function, 
	# Because http2 channels handle multiple responses concurrently
	# Practically USE http2_producer_fifo
	
	def __init__ (self):
		self.l = deque ()
		self.r = deque ()
		self.has_None = False
	
	def working (self):
		return (self.has_None or self.l or self.r) and 1 or 0
	
	def __len__ (self):
		if self.l:
			if not hasattr (self.l [0], 'ready'):
				return 1
			if not self.l [0].ready ():
				item = self.l.popleft ()
				self.r.append (item)
		
		if self.l:
			return 1
	
		if self.r:
			for i in range (len (self.r)):				
				if self.r [0].ready ():
					self.l.append (self.r.popleft ())
					return 1	
				self.r.rotate (1)				
		
		if self.has_None and not self.l and not self.r:
			# for return None
			self.l.append (None)
			self.has_None = False
			return 1
			
		return 0
			
	def __getitem__(self, index):
		return self.l [index]			
				
	def __setitem__(self, index, item):
		if index == 0:
			self.l.appendleft (item)
		else:
			self.l.append (item)
		
	def __delitem__ (self, index):
		del self.l [index]
	
	def append (self, item):
		self.insert (-1, item)
	
	def appendleft (self, item):
		self.insert (0, item)
	
	def insert_into (self, lst, index, item):		
		if index == 0:
			lst.appendleft (item)
		else:
			lst.append (item)	
		
	def insert (self, index, item):
		if item is None:
			self.has_None = True
			return			
		if self.has_None:
			return # deny adding			
		if hasattr (item, 'ready'):
			return self.r.append (item)		
		self.insert_into (self.l, index, item)
		
	def clear (self):
		self.l.clear ()
		self.r.clear ()
	
	
class ready_producer_ts_fifo (ready_producer_fifo):
	def __init__ (self):
		ready_producer_fifo.__init__ (self)
		self._lock = threading.Lock ()
	
	def working (self):
		with self._lock:
			return ready_producer_fifo.working (self)
		
	def __len__ (self):
		with self._lock:
			return ready_producer_fifo.__len__ (self)
			
	def __getitem__(self, index):
		with self._lock:
			return ready_producer_fifo.__getitem__ (self, index)
		
	def __setitem__(self, index, item):
		with self._lock:
			ready_producer_fifo.__setitem__ (self, index, item)
		
	def __delitem__ (self, index):
		with self._lock:
			ready_producer_fifo.__delitem__ (self, index)
	
	def clear (self):
		with self._lock:
			ready_producer_fifo.clear (self)
			
	def insert (self, index, item):
		with self._lock:
			ready_producer_fifo.insert (self, index, item)


from . import asynredis	
import struct
import socket
from pymongo import auth, message, helpers
from bson.codec_options import CodecOptions
from bson.son import SON

DEBUG = True
MAXINT = 2 ** (struct.Struct('i').size * 8 - 1) - 1

class MongoDBError (Exception):
	pass
	
class MongoMessageHandler:
	def __init__ (self, asyncon):
		self.asyncon = asyncon
		self.dbname = asyncon.dbname				
		self.length = -1
		self.request_id = 0
	
	def close (self):	
		self.asyncon = None

	def send_message (self, message):
		try:
			self.request_id, data, max_bson_size = message
		except ValueError:			
			self.request_id, data = message		
		self.asyncon.push (data)
		self.asyncon.set_terminator (16)
		
	def _parse_header (self, header):
		"""
		This method and below _parse_response() is based on 
		asyncmongo: An asynchronous library for accessing mongo with tornado.ioloop
		Author: Bitly
		URL: https://github.com/bitly/asyncmongo
		"""
		length = int(struct.unpack("<i", header[:4])[0])
		request_id = struct.unpack("<i", header[8:12])[0]
		assert request_id == self.request_id, \
			"ids don't match %r %r" % (self.request_id, request_id)
		operation = 1 # who knows why
		assert operation == struct.unpack("<i", header[12:])[0]
		self.length = length - 16		
		self.asyncon.set_terminator (self.length)
		
	def _parse_response (self, response):
		self.length = -1
		self.asyncon.set_terminator (16)
		request_id = self.request_id
		self.request_id = 0
		try:
			response = helpers._unpack_response(response, request_id)
		except:			
			self.asyncon.handle_error ()			
		if response and response['data'] and response['data'][0].get('err') and response['data'][0].get('code'):
			raise MongoDBError ("[%s] %s" % (response['data'][0]['code'], response['data'][0]['err']))
		else:
			self.asyncon.handle_response (response)		
		
	
class AsynConnect (asynredis.AsynConnect):	
	def __init__ (self, address, params = None, lock = None, logger = None):
		asynredis.AsynConnect.__init__ (self, address, params, lock, logger)
		self.mgh = MongoMessageHandler (self)
		self.codec_option = CodecOptions (SON)
		self._state = None
		
	def handle_connect (self):
		self.set_event_time ()		
		if self.user:
			self._state = "nonce"
			msg = message.query(0, "%s.$cmd" % self.dbname,	0, 1, SON({'getnonce': 1}),SON({}))
			self.mgh.send_message (msg)
			
	def push_command (self, msg):
		self.set_event_time ()
		self.mgh.send_message (msg)
	
	def get_full_colname (self, colname):	
		return "%s.%s" % (self.dbname, colname)

	def is_response_expected (self):		
		if self.last_command == "kill_cursors" and not self.producer_fifo:
			self.handle_response (None)
			
	def handle_write (self):
		asynredis.AsynConnect.handle_write (self)	
		self.is_response_expected ()
			
	def push (self, data):
		asynredis.AsynConnect.push (self, data)	
		self.is_response_expected ()
	
	def found_terminator (self):		
		if self.user and self._state is not None:
			if self._state == "nonce":
				self._state = "finish"
				response = self.fetchall ()
				try:
					nonce = response['data'][0]['nonce']				
					key = auth._auth_key(nonce, self.dbuser, self.dbpass)
				except:
					raise AuthenticationError ("Authentication Error")
				msg = message.query (
					0, "%s.$cmd" % self.pool._dbname, 0, 1,
					SON([('authenticate', 1), ('user', self.user), ('nonce', nonce), ('key', password)]), SON({})
				)
				return self.send_message (msg)
				
			elif self._state == "finish":
				self._state = None
				response = self.fetchall ()				
				try:
					assert response ['number_returned'] == 1
					response = response ['data'][0]
				except:
					raise AuthenticationError ("Authentication Error")
				if response.get("ok") != 1:				
					raise AuthenticationError (response.get("errmsg"))
			return	
	
		if self.mgh.length == -1:
			self.mgh._parse_header (self.data [-1])
		else:			
			self.mgh._parse_response (self.data [-1])
		self.data = []
		
	def handle_response (self, response):
		if response is not None: # if None end of session
			if self.last_command [:6] in ("insert", "upsert", "delete"):
				data = response ["data"][0]
				if data ["ok"] != 1.0:
					raise MongoDBError ("%(err)s (connectionId:%(connectionId)d syncMillis:%(syncMillis)d)" % data)
					
			self.response = response
			if not self.preserve_cursor:
				cursor_id = response.get ('cursor_id')
				if cursor_id != 0:
					self.response ['cursor_id'] = 0
					return self.kill_cursors ([cursor_id])

		self.has_result = True
		self.close_case_with_end_tran ()
		
	def fetchall (self):
		r = self.response
		self.response = None
		self.has_result = False
		return r
	
	def __query (self, colname, spec, offset = 0, limit = 1):
		# options, collection_name, num_to_skip, num_to_return, query, field_selector, opts, check_keys
		if limit == -1: limit = MAXINT
		msg = message.query (0, self.get_full_colname (colname), offset, limit, spec, None, self.codec_option)
		self.push_command (msg)
	
	def find (self, colname, spec, offset = 0, limit = 1):
		self.preserve_cursor = False
		self.__query (colname, spec, offset, limit)
	
	def findone (self, colname, spec):
		self.preserve_cursor = False
		self.__query (colname, spec, 0, 1)	
	
	def findall (self, colname, spec):
		self.preserve_cursor = False
		self.__query (colname, spec, 0, -1)	
	
	def findkc (self, colname, spec, offset = 0, limit = 1):
		self.preserve_cursor = True
		self.__query (colname, spec, offset, limit)
		
	def get_more (self, colname, cursor_id, num_to_return):
		self.preserve_cursor = True
		msg = message.get_more (self.get_full_colname (colname), num_to_return, cursor_id)
		self.push_command (msg)
	
	def kill_cursors (self, cursor_ids):
		self.last_command = "kill_cursors"
		if type (cursor_ids) not in (list, tuple): cursor_ids = [cursor_ids]
		msg = message.kill_cursors (cursor_ids)
		self.push_command (msg)
	
	def insert (self, colname, docs, continue_on_error = 0):
		#collection_name, docs, check_keys, safe, last_error_args, continue_on_error, opts
		if type (docs) not in (list, tuple): docs = [docs]		
		msg = message.insert (self.get_full_colname (colname), docs, False, 1, (), continue_on_error, self.codec_option)	
		self.push_command (msg)
	
	def __update (self, colname, spec, doc, multi = 1, upsert = 0):
		#collection_name, upsert, multi, spec, doc, safe, last_error_args, check_keys, opts
		msg = message.update (self.get_full_colname (colname), upsert, multi, spec, doc, 1, (), False, self.codec_option)
		self.push_command (msg)
	
	def update (self, colname, spec, doc):
		self.__update (colname, spec, doc, 1, 0)
		
	def updateone (self, colname, spec, doc):
		self.__update (colname, spec, doc, 0, 0)
		
	def upsert (self, colname, spec, doc):
		self.__update (colname, spec, doc, 1, 1)
	
	def upsertone (self, colname, spec, doc):
		self.__update (colname, spec, doc, 0, 1)
		
	def delete (self, colname, spec, flag = 0):
		#collection_name, spec, safe, last_error_args, opts, flags=0
		msg = message.delete (col, args [0], 1, (), self.codec_option, flags)		
		self.push_command (msg)
	
	def begin_tran (self, request):
		asynredis.AsynConnect.begin_tran (self, request)
		self.response = None
		self.last_command = request.method.lower ()
		self.preserve_cursor = False
		self.data = []
	
	REQ_OPS_OF_WIRE_PROTOCOL = (
		"find", "findone", "findall", 
		"kfind", "get_more", "kill_cursors",
		"insert", "delete",
		"update", "upsert", "updateone", "upsertone"		
	)
	def execute (self, request):
		# SHOULD push before adding to map, otherwise raised threading collision
		command = request.method.lower ()
		if command not in self.REQ_OPS_OF_WIRE_PROTOCOL:
			raise NotImplementedError ("Command %s Not Impemented" % command)		
		self.begin_tran (request)		
		getattr (self, command) (*request.params)
		
		self.set_terminator (16)
		if not self.connected:
			self.connect ()
		else:
			self.add_channel ()
		


import asynchat, asyncore
import re, os, sys
import ssl
import socket
import time
import zlib
from warnings import warn
from errno import ECONNRESET, ENOTCONN, ESHUTDOWN, ECONNABORTED, EWOULDBLOCK
if os.name == "nt":
	from errno import WSAENOTCONN
import select
import threading
from . import adns
from aquests.protocols.http2 import H2_PROTOCOLS
from aquests.lib.athreads.fifo import ready_producer_fifo
from aquests.lib.ssl_ import resolve_cert_reqs, resolve_ssl_version, create_urllib3_context
from collections import deque

NPN_PROTOCOL = 'h2'
H2_NPN_PROTOCOLS = [NPN_PROTOCOL, 'h2-16', 'h2-15', 'h2-14']
DEBUG = False
	
class SocketPanic (Exception): pass
class TimeOut (Exception): pass

class AsynConnect (asynchat.async_chat):
	ac_in_buffer_size = 65535
	ac_out_buffer_size = 65535
	zombie_timeout = 10
	keep_alive_timeout = 10
	network_delay_timeout = 10
	request_count = 0
	active = 0
	proxy = False
	fifo_class = deque
	
	def __init__ (self, address, lock = None, logger = None):
		self.address = address
		self.lock = lock
		self.logger = logger
		self._cv = threading.Condition ()		
		self.__sendlock = None
		self.__no_more_request = False
		self.set_event_time ()
		self.proxy = False
		self.proxy_client = False
		self.handler = None
		self.initialize_connection ()
		#asynchat.async_chat.__init__ (self)
		self.ac_in_buffer = b''
		self.incoming = []
		self.producer_fifo = self.fifo_class ()
		asyncore.dispatcher.__init__(self)
	
	def close (self):
		if self._closed:
			return
			
		if self.socket:
			# self.socket is still None, when DNS not found
			asynchat.async_chat.close (self)
		
		# re-init asychat
		self.ac_in_buffer = b''
		self.incoming = []
		self.producer_fifo.clear()		
		self._proto = None
		
		if self.handler and self.errcode:		
			self.handler.connection_closed (self.errcode, self.errmsg)
		
		self.set_active (False)
		self.handler = None
		self._closed = True
		
		if not self.proxy_client:
			self.logger ("[info] .....socket %s:%d has been closed" % self.address)
		
	def end_tran (self):
		self.del_channel ()
		self.handler = None
		self.set_active (False)
				
	def use_sendlock (self):
		self.__sendlock = threading.Lock ()
		self.initiate_send = self._initiate_send_ts
		
	def _initiate_send_ts (self):
		with self.__sendlock:
			return asynchat.async_chat.initiate_send (self)
				
	def get_proto (self):
		with self.lock:
			p = self._proto
		return p
	
	def set_proto (self, proto):
		with self.lock:
			self._proto = proto		
				
	def get_history (self):
		return self.__history
				
	def initialize_connection (self):		
		self._closed = False
		self._raised_ENOTCONN = 0 # for win32
		self.__history = []
		self._proto = None
		self._handshaking = False
		self._handshaked = False		
		
		self.established = False		
		self.upgraded = False		
		
	def set_event_time (self):
		self.event_time = time.time ()
	
	def is_proxy (self):
		return self.proxy
			
	def log (self, msg, logtype):
		if self.handler is not None and hasattr (self.handler, "log"):
			self.handler.log (msg, logtype)
		elif self.logger:
			self.logger (msg, logtype)
		else:
			warn ("No logger")
			
	def trace (self):
		if self.handler is not None and hasattr (self.handler, "trace"):
			self.handler.trace ()
		elif self.logger:
			self.logger.trace ()
		else:
			warn ("No logger for traceback")
				
	def duplicate (self):
		return self.__class__ (self.address, self.lock, self.logger)
		
	def clean_shutdown_control (self, phase, time_in_this_phase):	
		self.__no_more_request = True
		if self.isactive () or (self.handler and self.handler.working ()):
			return 1
		else:				
			self.handle_close ()
			self.__no_more_request = False
			return 0
	
	def is_channel_in_map (self, map = None):
		if map is None:
			map = self._map
		return self._fileno in map
		
	def set_active (self, flag, nolock = False):
		if flag:
			flag = time.time ()
		else:
			flag = 0
			
		if nolock or self.lock is None:
			self.active = flag
			return
			
		self.lock.acquire ()
		self.active = flag
		self.request_count += 1
		self.lock.release ()
	
	def get_active (self, nolock = False):
		if nolock or self.lock is None:
			return self.active
		self.lock.acquire ()
		active = self.active
		self.lock.release ()	
		return active
	
	def isactive (self):	
		return self.get_active () > 0
	
	def isconnected (self):
		with self.lock:
			r = self.connected
		return r	
		
	def get_request_count (self):	
		return self.request_count
	
	def add_channel (self, map = None):
		self.zombie_timeout =  self.network_delay_timeout
		self._fileno = self.socket.fileno ()
		return asynchat.async_chat.add_channel (self, map)		
		
	def del_channel (self, map = None):
		asynchat.async_chat.del_channel (self, map)
		# make default and sometimes reset server'stimeout	
		self.zombie_timeout =  self.keep_alive_timeout
				
	def create_socket (self, family, type):
		self.family_and_type = family, type
		sock = socket.socket (family, type)
		sock.setblocking (0)
		self.set_socket (sock)
	
	def connect (self):
		if adns.query:
			adns.query (self.address [0], "A", callback = self.continue_connect)
		else:
			# no adns query
			self.continue_connect (True)
		
	def continue_connect (self, answer = None):
		if not answer:
			self.log ("DNS not found - %s" % self.address [0], "error")
			return self.handle_close (704, "DNS Not Found")
		
		self.initialize_connection ()
		self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
		
		try:
			if not adns.query:				
				asynchat.async_chat.connect (self, self.address)
				
			else:	
				res = adns.get (self.address [0], "A")	
				if res:
					ip = res [-1]["data"]
					if ip:
						asynchat.async_chat.connect (self, (ip, self.address [1]))																		
				else:
					self.handle_close (704, "DNS Not Found")
					
		except:	
			self.handle_error ()
	
	def recv (self, buffer_size):		
		self.set_event_time ()
		try:
			data = self.socket.recv (buffer_size)
			#print ("+", len (data), end = ' ')
			if not data:
				self.handle_close (700, "Connection closed unexpectedly in recv")
				return b''
			else:
				return data		
		except socket.error as why:
			if why.errno in asyncore._DISCONNECTED:
				self.handle_close (700, "Connection closed unexpectedly in recv")
				return b''				
			else:
				raise
	
	def send (self, data):
		self.set_event_time ()
		#print ("====SEND", data)
		try:
			return self.socket.send (data)
		except socket.error as why:
			if why.errno == EWOULDBLOCK:
				return 0				
			elif why.errno in asyncore._DISCONNECTED:
				#print (">>>>>>>>>> why.errno == asyncore.ENOTCONN", why.errno == asyncore.ENOTCONN)
				if os.name == "nt" and why.errno == asyncore.ENOTCONN:
					# winsock sometimes raise ENOTCONN and sometimes recovered.
					# Found this error at http://file.hungryboarder.com:8080/HBAdManager/pa.html?siteId=hboarder&zoneId=S-2
					if self._raised_ENOTCONN <= 3:
						self._raised_ENOTCONN += 1
						return 0
					else:
						self._raised_ENOTCONN = 0
				self.handle_close (700, "Connection closed unexpectedly in send")
				return 0
			else:
				raise
	
	def close_if_over_keep_live (self):
		if time.time () - self.event_time > self.zombie_timeout:
			self.disconnect ()
	
	def set_zombie_timeout (self, timeout = 10):
		self.zombie_timeout = timeout
		
	def set_timeout (self, a = 10, b = 10):
		self.set_keep_alive_timeout (a)
		self.set_network_delay_timeout (b)
		
	def set_network_delay_timeout (self, timeout = 10):
		self.network_delay_timeout = timeout
	
	def set_keep_alive_timeout (self, timeout = 10):
		self.keep_alive_timeout = timeout
		
	def handle_connect (self):
		if hasattr (self.handler, "has_been_connected"):		
			self.handler.has_been_connected ()
			
	def handle_expt (self):
		self.logger ("socket panic", "fail")
		self.handle_close (703, "Socket Panic")
	
	def handle_error (self):
		self.trace ()
		self.handle_close(701, "Exception")
	
	def handle_timeout (self):
		self.log ("socket timeout", "fail")
		self.handle_close (702, "Zombie Timeout")
		
	def handle_expt_event(self):
		err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
		if err != 0:
			self.log ("Socket Error %d Occurred" % err, "warn")
			self.handle_close (706, "Socket %d Error" % err)
		else:
			self.handle_expt ()
	
	def maintern (self, object_timeout):
		if time.time () - self.event_time > object_timeout:
			if self.handler and hasattr (self.handler, "enter_shutdown_process"):
				self.handler.enter_shutdown ()
				return True
			else:	
				self.disconnect ()
				return True		
		return False
		
	# proxy POST need no init_send
	def push (self, thing, init_send = True):
		if type (thing) is bytes:
			asynchat.async_chat.push (self, thing)
		else:
			self.push_with_producer (thing, init_send)	
		
	def push_with_producer (self, producer, init_send = True):
		self.producer_fifo.append (producer)
		if init_send:
			self.initiate_send ()
	
	def handle_abort (self):
		self.handler = None
		self.close ()
		
	def handle_close (self, code = 700, msg = "Disconnected by Server"):
		self.errcode = code
		self.errmsg = msg
		self.close ()
							
	def collect_incoming_data (self, data):
		if not self.handler:
			self.logger ("recv data but no hander, droping data %d" % len (data), "warn")
			self.disconnect ()
			return
		self.handler.collect_incoming_data (data)
	
	def found_terminator (self):
		if not self.handler:
			self.logger ("found terminator but no handler", "warn")
			self.disconnect ()
			return # already closed
		self.handler.found_terminator ()
	
	def disconnect (self):
		# no error
		self.handle_close (0, "")
	
	def reconnect (self):
		self.disconnect ()
		self.connect ()
	
	def set_proxy (self, flag = True):
		self.proxy = flag
		
	def set_proxy_client (self, flag = True):
		self.proxy_client = flag
						
	def begin_tran (self, handler):
		if self.__no_more_request:
			return self.handle_close (705, "Entered Shutdown Process")
		self.errcode = 0
		self.errmsg = ""
		
		self.handler = handler
		self.set_event_time ()
		self.proxy_client = False
		
		if self.connected:
			self.close_if_over_keep_live () # check keep-alive
		
		# IMP: call add_channel () AFTER push()	otherwise threading issue will be raised
		try:
			if self.connected:
				#should keep order but it seems meaningless? LOOK LATER with initiate_send
				#self.initiate_send ()
				self.add_channel ()
			else:
				self.connect ()
		except:
			self.handle_error ()


class AsynSSLConnect (AsynConnect):	
	ac_negotiate_http2 = True
	
	def negotiate_http2 (self, flag):
		self.ac_negotiate_http2 = flag
		
	def handshake (self):
		if not self._handshaking:
			err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
			if err != 0:
				raise socket.error(err, _strerror(err))								
			ssl_context = create_urllib3_context(ssl_version=resolve_ssl_version(None), cert_reqs=resolve_cert_reqs(None))
			if self.ac_negotiate_http2:
				try: ssl_context.set_alpn_protocols (H2_PROTOCOLS)
				except AttributeError: ssl_context.set_npn_protocols (H2_PROTOCOLS)								
			self.socket = ssl_context.wrap_socket (self.socket, do_handshake_on_connect = False, server_hostname = self.address [0])			
			self._handshaking = True
			
		try:
			self.socket.do_handshake ()
		except ssl.SSLError as why:
			if why.args [0] in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE):
				return False
			raise ssl.SSLError(why)
		
		try: self._proto = self.socket.selected_alpn_protocol()
		except (AttributeError, NotImplementedError): 
			try: self._proto = self.socket.selected_npn_protocol()
			except (AttributeError, NotImplementedError): pass

		self._handshaked = True
		return True
							
	def handle_connect_event (self):	
		if not self._handshaked and not self.handshake ():
			return
		# handshaking done
		self.handle_connect()
		self.connected = True
		
	def recv (self, buffer_size):
		self.set_event_time ()
		try:
			data = self.socket.recv (buffer_size)			
			if not data:				
				self.handle_close (700, "Connection closed unexpectedly")
				return b''
			else:				
				return data
			
		except ssl.SSLError as why:
			if why.errno == ssl.SSL_ERROR_WANT_READ:
				try: 
					raise BlockingIOError				
				except NameError:
					raise socket.error (EWOULDBLOCK)
													
			# closed connection
			elif why.errno in (ssl.SSL_ERROR_ZERO_RETURN, ssl.SSL_ERROR_EOF):
				self.handle_close (700, "Connection closed by SSL_ERROR_ZERO_RETURN or SSL_ERROR_EOF")
				return b''
				
			else:
				raise

	def send (self, data):
		self.set_event_time ()
		try:
			return self.socket.send (data)			

		except ssl.SSLError as why:
			if why.errno == ssl.SSL_ERROR_WANT_WRITE:
				return 0
			elif why.errno == ssl.SSL_ERROR_ZERO_RETURN:				
				self.handle_close (700, "Connection closed by SSL_ERROR_ZERO_RETURN")
				return 0
			else:
				raise


class AsynSSLProxyConnect (AsynSSLConnect, AsynConnect):
	ac_out_buffer_size = 65536
	ac_in_buffer_size = 65536
	
	def handle_connect_event (self):
		if self.established:
			AsynSSLConnect.handle_connect_event (self)
		else:	
			AsynConnect.handle_connect_event (self)
	
	def recv (self, buffer_size):		
		if self._handshaked or self._handshaking:
			return AsynSSLConnect.recv (self, buffer_size)				
		else:
			return AsynConnect.recv (self, buffer_size)
			
	def send (self, data):		
		if self._handshaked or self._handshaking:
			return AsynSSLConnect.send (self, data)
		else:
			return AsynConnect.send (self, data)

	


import aquests
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib
import struct
import base64
import json
try: 
	from urllib.parse import urlparse, quote
except ImportError:
	from urllib import quote
	from urlparse import urlparse
				
from aquests.lib import producers, strutil

class XMLRPCRequest:	
	user_agent = "Mozilla/5.0 (compatible; Aquests/%s.%s)" % aquests.version_info [:2]
	initial_http_version = "1.1"
			
	def __init__ (self, uri, method, params = (), headers = None, auth = None, logger = None, meta = {}, http_version = None):
		# mount point generalizing, otherwise some servers reponse 301
		self.uri = uri
		self.method = method
		self.params = params		
		self.auth = (auth and type (auth) is not tuple and tuple (auth.split (":", 1)) or auth)
		self.logger = logger
		self.meta = meta
		self.http_version = http_version
		self.content_length = 0
		self.address, self.path = self.split (uri)
		self.__xmlrpc_serialized = False
		self.headers = {
			"Accept": "*/*",
			"Accept-Encoding": "gzip"
		}
		if headers:
			for k, v in headers.items ():
				n = k.lower ()
				if n in ("accept-encoding	", "content-length", "connection"):
					# reanalyze
					continue					
				self.headers [k] = v			
		self.payload = self.serialize ()
		
	def set_content_length (self, length):
		self.content_length = length
	
	def get_content_length (self):
		return self.content_length
				
	def build_header (self):
		if self.get_header ("host") is None:
			address = self.get_address ()
			if address [1] in (80, 443):			
				self.headers ["Host"] = "%s" % address [0]
			else:
				self.headers ["Host"] = "%s:%d" % address
		
		if self.get_header ("user-agent") is None:
			self.headers ["User-Agent"] = self.user_agent
		
	def get_cache_key (self):
		if len (self.payload) > 4096:
			return None
		return "%s:%s%s/%s?%s" % (
			self.address [0], self.address [1],
			self.path, self.method, self.payload
		)
		
	def xmlrpc_serialized (self):
		return self.__xmlrpc_serialized
		
	def set_address (self, address):
		self.address = address
		
	def get_address (self):
		return self.address
		
	def get_method (self):
		return "POST"
		
	def split (self, uri):
		if uri.find ("://") == -1:
			return None, uri
			
		scheme, address, script, params, qs, fragment = urlparse (uri)
		if not script: script = "/"
		path = script
		if params: path += ";" + params
		if qs: path += "?" + qs
		
		try: 
			host, port = address.split (":", 1)
			port = int (port)
		except ValueError:
			host = address
			if scheme in ("http", "ws"):
				port = 80
			else:
				port = 443	
		
		return (host, port)	, path
		
	def serialize (self):
		self.__xmlrpc_serialized = True		
		if self.uri [-1] != "/":
			self.uri += "/"
			self.path += "/"
		data = xmlrpclib.dumps (self.params, self.method, allow_none = 1).encode ("utf8")
		self.headers ["Content-Type"] = "text/xml"
		cl = len (data)
		self.headers ["Content-Length"] = cl
		self.set_content_length (cl)
		return data
	
	def get_auth (self):
		return self.auth
		
	def get_payload (self):
		return self.payload
	get_data = get_payload	
	
	def get_header (self, k, with_key = False):
		if self.headers:
			k = k.lower ()
			for n, v in self.headers.items ():
				if n.lower () == k:
					if with_key:
						return n, v
					return v
					
		if with_key:
			return None, None		
		
	def get_headers (self):
		self.build_header ()
		return list (self.headers.items ())


	
class HTTPRequest (XMLRPCRequest):		
	def get_method (self):
		return self.method.upper ()
	
	def get_cache_key (self):
		if len (self.payload) > 4096:
			return None			
		return "%s:%s/%s?%s" % (
			self.address [0], self.address [1], self.path, self.payload
		)
	
	def to_bytes (self, data, set_content_length = True):
		if type (self.params) is not bytes:			
			if strutil.is_encodable (self.params):
				data = self.params.encoding ("utf8")			
			
		if set_content_length:
			# when only bytes type, in case proxy_request this value will be just bool type
			try:
				cl = len (data)				
			except TypeError:
				pass
			else:
				self.headers ["Content-Length"] = cl
				self.set_content_length (cl)	
				
		return data
	
	def urlencode (self, to_bytes = True):
		fm = []
		for k, v in list(self.params.items ()):			
			fm.append ("%s=%s" % (quote (k), quote (v)))				
		if to_bytes:	
			return "&".join (fm).encode ("utf8")
		return "&".join (fm)
	
	def nvpencode (self):
		fm = []
		for k, v in list(self.params.items ()):
			v = v.encode ("utf8")
			fm.append (k.encode ("utf8") + b"[" + str (len (v)).encode ("utf8") + b"]=" + v)
		return b"&".join (fm)
									
	def serialize (self):
		if not self.params:
			if self.get_method () in ("POST", "PUT"):
				self.headers ["Content-Length"] = 0
			return b""
		
		# formdata type can be string, dict, boolean
		if self.get_method () in ("GET", "DELETE"):
			if type (self.params) is dict:
				params = self.urlencode (to_bytes = False)
			else:
				params = self.params
			self.uri += "?" + params
			self.path += "?" + params
			self.params = None
			return b""
		
		data = self.params
		header_name, content_type = self.get_header ("content-type", True)
		
		if not content_type:
			content_type = "application/json"
						
		if type (self.params) is dict:			
			if content_type == "application/json":
				data = json.dumps (self.params).encode ("utf8")				
			elif content_type == "application/x-www-form-urlencoded":
				data = self.urlencode ()				
			elif content_type == "text/namevalue":
				data = self.nvpencode ()				
		self.headers ["Content-Type"] = content_type
		return self.to_bytes (data)
		
		
class HTTPMultipartRequest (HTTPRequest):
	boundary = "-------------------Skitai-%s.%s-a1a80da4-ca3d-11e6-b245-001b216d6e71" % aquests.version_info [:2]
		
	def __init__ (self, uri, method, params = {}, headers = None, auth = None, logger = None, meta = {}):
		HTTPRequest.__init__ (self, uri, method, params, headers, auth, logger, meta)
		if type (self.params) is bytes:
			self.find_boundary ()
	
	def get_cache_key (self):
		return None
		
	def get_method (self):
		return "POST"
			
	def find_boundary (self):
		s = self.params.find (b"\r\n")
		if s == -1:
			raise ValueError("Boundary Not Found")
		b = self.params [:s]			
		if b [:2] != b"--":
			raise ValueError("invalid multipart/form-data")
		self.boundary = b [2:s]
		
	def serialize (self):
		self.headers ["Content-Type"] = "multipart/form-data; boundary=" + self.boundary
		if type (self.params) is dict:
			p = producers.multipart_producer (self.params, self.boundary)
			cl = p.get_content_length ()
			self.headers ["Content-Length"] = cl
			self.set_content_length (cl)
			return p
		return self.to_bytes (self.params)		
	
